@@ -268,22 +268,22 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
         create_circuit_test(
             "compiler_simp",
             ablated.clone(),
-            compiler_simp(&ablated, context),
+            compiler_simp(ablated, context),
         )
         .unwrap();
     }
-    let circuit = timed!(compiler_simp(&circuit, context), 10, print_timings);
+    let circuit = timed!(compiler_simp(circuit, context), 10, print_timings);
     // originally tried push_down_index in compiler_simp, but that produced worse circuits
     // for unknown reasons, maybe i'll investigate
     let circuit = timed!(
         apply_fn_until_same(&circuit, |x| deep_push_down_index(
-            compiler_simp(x, context),
+            compiler_simp(x.clone(), context),
             None
         )),
         10,
         print_timings
     );
-    let circuit = timed!(compiler_simp(&circuit, context), 10, print_timings);
+    let circuit = timed!(compiler_simp(circuit, context), 10, print_timings);
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         println!("Simplified Circuit");
@@ -291,8 +291,8 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     }
     let circuit = timed!(
         apply_fn_until_same(&circuit, |x| {
-            let distributed = deep_maybe_distribute(x, context);
-            compiler_simp(&distributed, context)
+            let distributed = deep_maybe_distribute(x.clone(), context);
+            compiler_simp(distributed, context)
         }),
         10,
         print_timings
@@ -421,7 +421,7 @@ pub fn deep_optimize_module_specs(
     context: &mut OptimizationContext,
 ) -> CircuitRc {
     deep_map_op_context(
-        &circuit,
+        circuit.clone(),
         &|circuit, context| {
             circuit.as_module_node().map(|mn| {
                 ModuleNode::nrc(
@@ -434,5 +434,5 @@ pub fn deep_optimize_module_specs(
         context,
         &mut Default::default(),
     )
-    .unwrap_or(circuit.clone())
+    .unwrap_or(circuit)
 }
