@@ -5,6 +5,7 @@ use crate::py_types::{
 use crate::pyo3_prelude::*;
 use crate::smallvec::Sv;
 use crate::sv;
+use crate::util::cumsum;
 use itertools::{zip, Itertools};
 use macro_rules_attribute::apply;
 use num_bigint::BigUint;
@@ -97,7 +98,7 @@ impl CircuitNode for Einsum {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
             self.args
@@ -105,7 +106,7 @@ impl CircuitNode for Einsum {
                 .enumerate()
                 .map(
                     move |(i, (circ, axes))| -> Result<_, CircuitConstructionError> {
-                        Ok((f(i, circ)?, axes.clone()))
+                        Ok((f(i, circ.clone())?, axes.clone()))
                     },
                 )
                 .collect::<Result<_, _>>()?,
@@ -229,45 +230,6 @@ impl Einsum {
         )
     }
 
-    pub fn get_spec(&self) -> EinsumSpec {
-        let to_usize = |vals: &[u8]| -> Vec<usize> { vals.iter().map(|&x| x as usize).collect() };
-        EinsumSpec {
-            input_ints: self.input_axes().map(|x| to_usize(x)).collect(),
-            output_ints: to_usize(&self.out_axes),
-            int_sizes: dict_to_list(
-                &self
-                    .shape_map()
-                    .unwrap()
-                    .iter()
-                    .map(|(key, val)| (*key as usize, *val))
-                    .collect(),
-                None,
-            ),
-        }
-    }
-
-    pub fn scalar_mul(node: CircuitRc, scalar: f64, name: Option<String>) -> Self {
-        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
-        Einsum::try_new(
-            vec![
-                (node, axes.clone()),
-                (ScalarConstant::new(scalar, sv![], None).rc(), sv![]),
-            ],
-            axes,
-            name,
-        )
-        .unwrap()
-    }
-
-    pub fn empty(name: Option<String>) -> Self {
-        Einsum::try_new(vec![], sv![], name).unwrap()
-    }
-
-    pub fn identity(node: CircuitRc, name: Option<String>) -> Self {
-        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
-        Einsum::try_new(vec![(node, axes.clone())], axes, name).unwrap()
-    }
-
     pub fn axes_in_input(&self) -> HashSet<EinInt> {
         // could be 256 bit bitmask
         self.input_axes().flatten().copied().collect()
@@ -430,6 +392,79 @@ impl Einsum {
             .map(|x| x + 1)
             .unwrap_or(0)
     }
+    pub fn get_spec(&self) -> EinsumSpec {
+        let to_usize = |vals: &[u8]| -> Vec<usize> { vals.iter().map(|&x| x as usize).collect() };
+        EinsumSpec {
+            input_ints: self.input_axes().map(|x| to_usize(x)).collect(),
+            output_ints: to_usize(&self.out_axes),
+            int_sizes: dict_to_list(
+                &self
+                    .shape_map()
+                    .unwrap()
+                    .iter()
+                    .map(|(key, val)| (*key as usize, *val))
+                    .collect(),
+                None,
+            ),
+        }
+    }
+    #[staticmethod]
+    pub fn scalar_mul(node: CircuitRc, scalar: f64, name: Option<String>) -> Self {
+        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
+        Einsum::try_new(
+            vec![
+                (node, axes.clone()),
+                (ScalarConstant::new(scalar, sv![], None).rc(), sv![]),
+            ],
+            axes,
+            name,
+        )
+        .unwrap()
+    }
+    #[staticmethod]
+    pub fn empty(name: Option<String>) -> Self {
+        Einsum::try_new(vec![], sv![], name).unwrap()
+    }
+
+    #[staticmethod]
+    pub fn identity(node: CircuitRc, name: Option<String>) -> Self {
+        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
+        Einsum::try_new(vec![(node, axes.clone())], axes, name).unwrap()
+    }
+
+    #[staticmethod]
+    pub fn new_outer_product(
+        nodes: Vec<CircuitRc>,
+        name: Option<String>,
+        out_axes_permutation: Option<Vec<usize>>,
+    ) -> Self {
+        let sections: Vec<_> = nodes.iter().map(|n| n.info().rank()).collect();
+        let starts = cumsum(&sections);
+        let mut out_axes: EinsumAxes = nodes
+            .iter()
+            .enumerate()
+            .flat_map(|(i, _n)| (starts[i] as u8..(starts[i] + sections[i]) as u8))
+            .collect();
+        if let Some(permutation) = out_axes_permutation {
+            let new_out_axes = permutation.iter().map(|i| out_axes[*i]).collect();
+            out_axes = new_out_axes;
+        }
+        Einsum::try_new(
+            nodes
+                .iter()
+                .enumerate()
+                .map(|(i, n)| {
+                    (
+                        n.clone(),
+                        (starts[i] as u8..(starts[i] + sections[i]) as u8).collect(),
+                    )
+                })
+                .collect(),
+            out_axes,
+            name,
+        )
+        .unwrap()
+    }
 }
 
 /// Add supports broadcasting
@@ -573,13 +608,13 @@ impl CircuitNode for Add {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
             self.nodes
                 .iter()
                 .enumerate()
-                .map(move |(i, circ)| f(i, circ))
+                .map(move |(i, circ)| f(i, circ.clone()))
                 .collect::<Result<Vec<_>, _>>()?,
             self.name.clone(),
         )
@@ -763,9 +798,13 @@ impl CircuitNode for Rearrange {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
-        Self::try_new(f(0, &self.node)?, self.spec.clone(), self.name.clone())
+        Self::try_new(
+            f(0, self.node.clone())?,
+            self.spec.clone(),
+            self.name.clone(),
+        )
     }
 
     fn eval_tensors(
@@ -951,9 +990,13 @@ impl CircuitNode for Index {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
-        Self::try_new(f(0, &self.node)?, self.index.clone(), self.name.clone())
+        Self::try_new(
+            f(0, self.node.clone())?,
+            self.index.clone(),
+            self.name.clone(),
+        )
     }
 
     fn eval_tensors(
@@ -1121,13 +1164,13 @@ impl CircuitNode for Concat {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
             self.nodes
                 .iter()
                 .enumerate()
-                .map(move |(i, circ)| f(i, circ))
+                .map(move |(i, circ)| f(i, circ.clone()))
                 .collect::<Result<Vec<_>, _>>()?,
             self.axis,
             self.name.clone(),
@@ -1316,10 +1359,10 @@ impl CircuitNode for Scatter {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
-            f(0, &self.node)?,
+            f(0, self.node.clone())?,
             self.index.clone(),
             self.info().shape.clone(),
             self.name.clone(),
@@ -1391,3 +1434,26 @@ pub fn flat_concat(circuits: Vec<CircuitRc>) -> Concat {
         .collect();
     Concat::try_new(flatteneds, 0, Some("flat_concat".to_owned())).unwrap()
 }
+
+#[pyfunction]
+pub fn flat_concat_back(circuits: Vec<CircuitRc>) -> (Concat, Vec<CircuitRc>) {
+    let flat = flat_concat(circuits.clone());
+    let sections = flat.get_sections();
+    let starts = cumsum(&sections);
+    (
+        flat.clone(),
+        zip(circuits, zip(starts, sections))
+            .map(|(c, (start, sec))| {
+                Rearrange::nrc(
+                    Index::nrc(
+                        flat.clone().rc(),
+                        TensorIndex(vec![TensorAxisIndex::new_plain_slice(start, start + sec)]),
+                        None,
+                    ),
+                    RearrangeSpec::unflatten(c.info().shape.clone()),
+                    None,
+                )
+            })
+            .collect(),
+    )
+}
