@@ -37,7 +37,7 @@ macro_rules! circuit_node_auto_leaf_impl {
         ) -> Result<Self, CircuitConstructionError>
         where
             CircuitConstructionError: From<E>,
-            F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+            F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
         {
             Ok(self.clone())
         }
