@@ -252,13 +252,13 @@ impl CircuitNode for GeneralFunction {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
             self.nodes
                 .iter()
                 .enumerate()
-                .map(move |(i, circ)| f(i, circ))
+                .map(move |(i, circ)| f(i, circ.clone()))
                 .collect::<Result<Vec<_>, _>>()?,
             self.spec.clone(),
             self.name.clone(),
