@@ -12,9 +12,10 @@ use crate::{
     util::AsOp,
 };
 pub use computational_node::{
-    flat_concat, Add, Concat, Einsum, EinsumAxes, Index, Rearrange, Scatter,
+    flat_concat, flat_concat_back, Add, Concat, Einsum, EinsumAxes, Index, Rearrange, Scatter,
 };
 pub use constant::{ArrayConstant, ScalarConstant, Symbol};
+pub use cumulant::Cumulant;
 pub use generalfunction::{GeneralFunction, GeneralFunctionSpec, PyGFSpecShapeGetter};
 pub use generalfunction_rewrite::{
     generalfunction_merge_inverses, generalfunction_special_case_simplification,
@@ -32,6 +33,7 @@ use std::{
     rc::Rc,
 };
 use uuid::uuid;
+pub use variable_nodes::{AutoTag, DiscreteVar, StoredCumulantVar};
 
 use macro_rules_attribute::apply;
 use std::fmt::Debug;
@@ -41,6 +43,7 @@ pub mod algebraic_rewrite;
 pub mod batching;
 pub mod canonicalize;
 pub mod circuit_manipulation;
+pub mod circuit_matchers;
 pub mod circuit_optimizer;
 pub mod circuit_utils;
 pub mod compiler_heuristics;
@@ -48,21 +51,25 @@ pub mod compiler_strip;
 mod computational_node;
 pub mod concat_rewrite;
 mod constant;
+pub mod cumulant;
 pub mod debugging;
 pub mod deep_rewrite;
 pub mod diag_rewrite;
 pub mod generalfunction;
 pub mod generalfunction_rewrite;
 pub mod module_nodes;
+pub mod module_rewrite;
 pub mod named_axes;
 pub mod nrc;
 mod parsing;
 pub mod print;
 pub mod py_circuit_items;
 mod repr;
+pub mod sampling;
 pub mod scatter_rewrite;
 pub mod scheduled_execution;
 pub mod scheduling_z3;
+pub mod variable_nodes;
 
 mod circuit_node_private {
     pub trait CircuitNodePrivate {
@@ -114,7 +121,7 @@ pub trait CircuitNode: CircuitNodePrivate {
     where
         Self: Sized,
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>;
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>;
 
     fn node_type_uuid(&self) -> [u8; 16];
 
@@ -174,7 +181,7 @@ pub trait CircuitNode: CircuitNodePrivate {
     where
         Self: Sized,
         CircuitConstructionError: From<E>,
-        F: FnMut(&'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(CircuitRc) -> Result<CircuitRc, E>,
     {
         self.map_children_enumerate(|_i, x| f(x))
     }
@@ -191,7 +198,7 @@ pub trait CircuitNode: CircuitNodePrivate {
     fn map_children_unwrap<'a, F>(&'a self, mut f: F) -> Self
     where
         Self: Sized,
-        F: FnMut(&'a Circuit) -> CircuitRc,
+        F: FnMut(CircuitRc) -> CircuitRc,
     {
         self.map_children(|x| Ok::<CircuitRc, CircuitConstructionError>(f(x)))
             .unwrap()
@@ -200,7 +207,7 @@ pub trait CircuitNode: CircuitNodePrivate {
     fn map_children_unwrap_enumerate<'a, F>(&'a self, mut f: F) -> Self
     where
         Self: Sized,
-        F: FnMut(usize, &'a Circuit) -> CircuitRc,
+        F: FnMut(usize, CircuitRc) -> CircuitRc,
     {
         self.map_children_enumerate(|i, x| Ok::<CircuitRc, CircuitConstructionError>(f(i, x)))
             .unwrap()
@@ -219,15 +226,15 @@ pub trait CircuitNode: CircuitNodePrivate {
     fn map_children_op<'a, F>(&'a self, mut f: F) -> Option<Self>
     where
         Self: Sized,
-        F: FnMut(&'a Circuit) -> Option<CircuitRc>,
+        F: FnMut(CircuitRc) -> Option<CircuitRc>,
     {
         let mut any_modified = false;
         let out = self.map_children_unwrap(|x| {
-            if let Some(new) = f(x) {
+            if let Some(new) = f(x.clone()) {
                 any_modified = true;
                 new
             } else {
-                x.clone().rc()
+                x
             }
         });
         if any_modified {
@@ -241,8 +248,7 @@ pub trait CircuitNode: CircuitNodePrivate {
     where
         Self: Sized,
     {
-        let shape = self.compute_shape();
-        self.info_mut().shape = shape.clone(); // set shape so methods to compute other info can use it
+        self.info_mut().shape = self.compute_shape(); // set shape so methods to compute other info can use it
 
         let mut hasher = self.compute_hash();
         self.info_mut().named_axes = self.compute_named_axes();
@@ -250,7 +256,7 @@ pub trait CircuitNode: CircuitNodePrivate {
         hasher.update(&self.node_type_uuid());
         hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes());
         for (axis, name) in &self.info().named_axes {
-            if *axis as usize >= shape.len() {
+            if *axis as usize >= self.info().shape.len() {
                 return Err(CircuitConstructionError::NamedAxisAboveRank {});
             }
             hasher.update(&[*axis]);
@@ -328,7 +334,7 @@ pub trait CircuitNodeUnion {
     where
         Self: Sized,
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>;
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>;
     fn variant_string(&self) -> String;
     fn node_type_uuid(&self) -> [u8; 16];
 }
@@ -459,7 +465,7 @@ macro_rules! define_circuit_union_impl {
             fn map_children_enumerate_impl<'a, F, E>(&'a self, f: F) -> Result<Self, $crate::circuit::CircuitConstructionError>
             where
                 $crate::circuit::CircuitConstructionError: From<E>,
-                F: FnMut(usize,&'a $crate::circuit::Circuit) -> Result<$crate::circuit::CircuitRc, E>,
+                F: FnMut(usize,$crate::circuit::CircuitRc) -> Result<$crate::circuit::CircuitRc, E>,
             {
                 match self {
                     $(
@@ -527,6 +533,10 @@ define_circuit!(
     Concat,
     Scatter,
     ModuleNode,
+    AutoTag,
+    DiscreteVar,
+    StoredCumulantVar,
+    Cumulant,
 );
 
 /// Define adhoc unions of different circuit types
@@ -560,6 +570,13 @@ macro_rules! define_circuit_union {
                 op.is_some()
             }
         }
+        paste::paste! {
+            #[pyfunction]
+            pub fn [<circuit_is_ $name:snake>](circuit:$crate::circuit::CircuitRc)->bool{
+                let op: ::std::option::Option<$name>=(**circuit).clone().into();
+                op.is_some()
+            }
+        }
     }
 }
 
@@ -577,6 +594,16 @@ define_circuit_union!(Leaf {
     ScalarConstant,
 });
 
+define_circuit_union!(Constant {
+    ArrayConstant,
+    ScalarConstant,
+});
+
+define_circuit_union!(Var {
+    StoredCumulantVar,
+    DiscreteVar,
+});
+
 // work around for fact that we can't implement foreign trait on constrained type
 #[macro_export]
 macro_rules! circuit_node_extra_impl {
@@ -690,7 +717,7 @@ impl<T: CircuitNodeUnion> CircuitNode for T {
     fn map_children_enumerate<'a, F, E>(&'a self, f: F) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         self.map_children_enumerate_impl(f)
     }
@@ -769,6 +796,28 @@ impl CachedCircuitInfo {
 
 #[derive(Error, Debug, Clone)]
 pub enum CircuitConstructionError {
+    #[error("DiscreteVar doesn't have leading 'samples' dim")]
+    DiscreteVarNoSamplesDim {},
+
+    #[error("DiscreteVar samples dim doesn't match probs, {node} vs {probs}")]
+    DiscreteVarWrongSamplesDim { node: usize, probs: usize },
+
+    #[error("DiscreteVar probs must be 1d with length matching samples axis 0, got probs of shape {shape:?}")]
+    DiscreteVarProbsMustBe1d { shape: Shape },
+
+    #[error("StoredCumulantVar needs first 2 cumulants specified")]
+    StoredCumulantVarNeedsMeanVariance {},
+
+    #[error("StoredCumulantVar invalid cumulant number {number}")]
+    StoredCumulantVarInvalidCumulantNumber { number: usize },
+
+    #[error("StoredCumulantVar cumulant number {cumulant_number} needs to be base shape, {base_shape:?} times cumulant number, got {cumulant_shape:?}")]
+    StoredCumulantVarCumulantWrongShape {
+        base_shape: Shape,
+        cumulant_shape: Shape,
+        cumulant_number: usize,
+    },
+
     #[error("len shape different from len axes (len axes: {len_axes}, len shape: {circuit_len_shape}, circuit name: {circuit_name:?})")]
     EinsumLenShapeDifferentFromAxes {
         circuit_name: Option<String>,
@@ -898,6 +947,9 @@ pub enum CircuitConstructionError {
         circuit: CircuitRc,
     },
 
+    #[error("Would need to batch multiple axes, only supports one")]
+    BatchingRequiresMultipleAxes {},
+
     #[error("ModuleNode got unknown keyword argument, {argument}")]
     ModuleNodeUnknownArgument { argument: String },
 
@@ -910,8 +962,19 @@ pub enum CircuitConstructionError {
     #[error("Parsing failed {string}")]
     ParsingFail { string: String },
 
-    #[error("Parsing invalid indentation, number of spaces {spaces}")]
-    ParsingInvalidIndentation { spaces: usize },
+    #[error("Expected UUID string, found  {string}")]
+    InvalidUuid { string: String },
+
+    #[error("Parsing wrong number of children, expected {expected} found {found}")]
+    ParsingWrongNumberChildren { expected: usize, found: usize },
+
+    #[error("Parsing invalid indentation, tab width {tab_width} num spaces {spaces} stack indentation {stack_indentation} stack top {stack_top:?}")]
+    ParsingInvalidIndentation {
+        spaces: usize,
+        tab_width: usize,
+        stack_indentation: usize,
+        stack_top: Option<String>,
+    },
 
     #[error("Parsing invalid circuit variant {v}")]
     ParsingInvalidVariant { v: String },
@@ -952,6 +1015,9 @@ pub enum CircuitConstructionError {
     #[error("Circuit by name not found: {name}")]
     CircuitRefNotFound { name: String },
 
+    #[error("Failed to construct equivalent explicitly computable circuit")]
+    NoEquivalentExplicitlyComputable {},
+
     #[error("python error {py_err:?}")]
     PythonError { py_err: Rc<PyErr> },
 }
@@ -983,6 +1049,8 @@ pub enum TensorEvalError {
         circ: TorchDeviceDtypeOp,
         passed: TorchDeviceDtypeOp,
     },
+    #[error("Failed to construct equivalent explicitly computable circuit")]
+    NoEquivalentExplicitlyComputable {},
 }
 
 impl From<TensorEvalError> for PyErr {
@@ -1003,17 +1071,17 @@ impl From<PyErr> for TensorEvalError {
 pub struct CircuitRc(Rc<Circuit>);
 
 pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
-    circuit.map_children_unwrap(&mut |child: &Circuit| {
+    circuit.map_children_unwrap(&mut |child: CircuitRc| {
         ScalarConstant::new(0.0, child.info().shape.clone(), child.name_cloned()).rc()
     })
 }
 
-pub fn evaluate_fn(circ: &Circuit) -> Result<Tensor, TensorEvalError> {
+pub fn evaluate_fn(circ: CircuitRc) -> Result<Tensor, TensorEvalError> {
     evaluate_fn_dtype_device(circ, Default::default())
 }
 
 pub fn evaluate_fn_dtype_device(
-    circ: &Circuit,
+    circ: CircuitRc,
     dtype_device: TorchDeviceDtypeOp,
 ) -> Result<Tensor, TensorEvalError> {
     let device_dtype = dtype_device
@@ -1026,116 +1094,116 @@ pub fn evaluate_fn_dtype_device(
         .unwrap_or_defaults();
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: Circuit) -> Result<Tensor, TensorEvalError> {
+    fn recurse(circ: CircuitRc) -> Result<Tensor, TensorEvalError> {
         let child_tensors: Result<Vec<Tensor>, TensorEvalError> =
-            circ.children().map(|x| recurse((**x).clone())).collect();
+            circ.children().map(recurse).collect();
         let child_tensors = child_tensors?;
 
         circ.eval_tensors(&child_tensors, &device_dtype)
     }
 
-    recurse(circ.clone())
+    recurse(circ)
 }
 
-pub fn deep_map<F>(circuit: &Circuit, f: F) -> Result<CircuitRc, CircuitConstructionError>
+pub fn deep_map<F>(circuit: CircuitRc, f: F) -> Result<CircuitRc, CircuitConstructionError>
 where
-    F: Fn(&Circuit) -> Result<CircuitRc, CircuitConstructionError>,
+    F: Fn(CircuitRc) -> Result<CircuitRc, CircuitConstructionError>,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
-        let inner_mapped = circ.map_children(&mut recurse)?;
-        f(&inner_mapped)
+    fn recurse(circ: CircuitRc) -> Result<CircuitRc, CircuitConstructionError> {
+        let inner_mapped = circ.map_children(&mut recurse)?.rc();
+        f(inner_mapped)
     }
     recurse(circuit)
 }
 
-pub fn deep_map_preorder<F>(circuit: &Circuit, f: F) -> Result<CircuitRc, CircuitConstructionError>
+pub fn deep_map_preorder<F>(circuit: CircuitRc, f: F) -> Result<CircuitRc, CircuitConstructionError>
 where
-    F: Fn(&Circuit) -> Result<CircuitRc, CircuitConstructionError>,
+    F: Fn(CircuitRc) -> Result<CircuitRc, CircuitConstructionError>,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
+    fn recurse(circ: CircuitRc) -> Result<CircuitRc, CircuitConstructionError> {
         f(circ)?.map_children(&mut recurse).map(|z| z.rc())
     }
     recurse(circuit)
 }
 
-pub fn deep_map_unwrap<F>(circuit: &Circuit, f: F) -> CircuitRc
+pub fn deep_map_unwrap<F>(circuit: CircuitRc, f: F) -> CircuitRc
 where
-    F: Fn(&Circuit) -> CircuitRc,
+    F: Fn(CircuitRc) -> CircuitRc,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> CircuitRc {
-        let inner_mapped = circ.map_children_unwrap(&mut recurse);
-        f(&inner_mapped)
+    fn recurse(circ: CircuitRc) -> CircuitRc {
+        let inner_mapped = circ.map_children_unwrap(&mut recurse).rc();
+        f(inner_mapped)
     }
     recurse(circuit)
 }
 
 /// this applies the function to parent before children, as opposed to deep_map_unwrap, which is children first
-pub fn deep_map_unwrap_preorder<F>(circuit: &Circuit, f: F) -> CircuitRc
+pub fn deep_map_unwrap_preorder<F>(circuit: CircuitRc, f: F) -> CircuitRc
 where
-    F: Fn(&Circuit) -> CircuitRc,
+    F: Fn(CircuitRc) -> CircuitRc,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> CircuitRc {
+    fn recurse(circ: CircuitRc) -> CircuitRc {
         f(circ).map_children_unwrap(&mut recurse).rc()
     }
     recurse(circuit)
 }
 
 /// visits each subcircuit that's in any circuit once! (pre-order traversal if it's a tree)
-pub fn visit_circuit<F>(circuit: &Circuit, mut f: F)
+pub fn visit_circuit<F>(circuit: CircuitRc, mut f: F)
 where
-    F: FnMut(&Circuit),
+    F: FnMut(CircuitRc),
 {
     let mut seen: HashSet<HashBytes> = HashSet::new();
 
-    fn recurse<F>(circ: &Circuit, seen: &mut HashSet<HashBytes>, f: &mut F)
+    fn recurse<F>(circ: CircuitRc, seen: &mut HashSet<HashBytes>, f: &mut F)
     where
-        F: FnMut(&Circuit),
+        F: FnMut(CircuitRc),
     {
         if !seen.contains(&circ.info().hash) {
             seen.insert(circ.info().hash);
-            f(circ);
+            f(circ.clone());
             for child in circ.children() {
-                recurse(&child, seen, f)
+                recurse(child, seen, f)
             }
         }
     }
     recurse(circuit, &mut seen, &mut f);
 }
 
-pub fn visit_circuit_with_parents<F>(circuit: &Circuit, mut f: F)
+pub fn visit_circuit_with_parents<F>(circuit: CircuitRc, mut f: F)
 where
-    F: FnMut(&Circuit, &Vec<CircuitRc>),
+    F: FnMut(CircuitRc, &Vec<CircuitRc>),
 {
-    let mut toposorted = toposort_circuit(circuit.clone().rc());
+    let mut toposorted = toposort_circuit(circuit);
     toposorted.reverse(); // its children first by default
 
     let mut parents: HashMap<CircuitRc, Vec<CircuitRc>> = HashMap::new();
     for (_i, sub) in toposorted.into_iter().enumerate() {
-        f(&sub, parents.get(&sub).unwrap_or(&vec![]));
+        f(sub.clone(), parents.get(&sub).unwrap_or(&vec![]));
         for child in sub.children() {
             parents.entry(child).or_insert(vec![]).push(sub.clone());
         }
     }
 }
 
-pub fn visit_circuit_with_parents_fallible<F, E>(circuit: &Circuit, mut f: F) -> Result<(), E>
+pub fn visit_circuit_with_parents_fallible<F, E>(circuit: CircuitRc, mut f: F) -> Result<(), E>
 where
-    F: FnMut(&Circuit, &Vec<CircuitRc>) -> Result<(), E>,
+    F: FnMut(CircuitRc, &Vec<CircuitRc>) -> Result<(), E>,
 {
-    let mut toposorted = toposort_circuit(circuit.clone().rc());
+    let mut toposorted = toposort_circuit(circuit);
     toposorted.reverse(); // its children first by default
 
     let mut parents: HashMap<CircuitRc, Vec<CircuitRc>> = HashMap::new();
     for (_i, sub) in toposorted.into_iter().enumerate() {
-        f(&sub, parents.get(&sub).unwrap_or(&vec![]))?;
+        f(sub.clone(), parents.get(&sub).unwrap_or(&vec![]))?;
         for child in sub.children() {
             parents.entry(child).or_insert(vec![]).push(sub.clone());
         }
@@ -1146,22 +1214,22 @@ where
 /// does not visit children of circuits where f fails. It does visit all children even if one fails
 /// even though this is more work than stopping on the first child that fails
 /// because it's semantically cleaner to not have to think about which children are first
-pub fn visit_circuit_fallable<F, E>(circuit: &Circuit, mut f: F) -> Result<(), E>
+pub fn visit_circuit_fallable<F, E>(circuit: CircuitRc, mut f: F) -> Result<(), E>
 where
-    F: FnMut(&Circuit) -> Result<(), E>,
+    F: FnMut(CircuitRc) -> Result<(), E>,
 {
     let mut seen: HashSet<HashBytes> = HashSet::new();
 
-    fn recurse<F, E>(circ: &Circuit, seen: &mut HashSet<HashBytes>, f: &mut F) -> Result<(), E>
+    fn recurse<F, E>(circ: CircuitRc, seen: &mut HashSet<HashBytes>, f: &mut F) -> Result<(), E>
     where
-        F: FnMut(&Circuit) -> Result<(), E>,
+        F: FnMut(CircuitRc) -> Result<(), E>,
     {
         if !seen.contains(&circ.info().hash) {
             seen.insert(circ.info().hash);
-            f(circ)?;
+            f(circ.clone())?;
 
             circ.children()
-                .map(|child| recurse(&child, seen, f))
+                .map(|child| recurse(child, seen, f))
                 .collect::<Vec<_>>() // intermediate collect causes all recurses to happen even if one errors
                 .into_iter()
                 .collect::<Result<Vec<_>, _>>()?;
@@ -1171,20 +1239,20 @@ where
     recurse(circuit, &mut seen, &mut f)
 }
 
-pub fn visit_circuit_postorder<F>(circuit: &Circuit, mut f: F)
+pub fn visit_circuit_postorder<F>(circuit: CircuitRc, mut f: F)
 where
-    F: FnMut(&Circuit),
+    F: FnMut(CircuitRc),
 {
     let mut seen: HashSet<HashBytes> = HashSet::new();
 
-    fn recurse<F>(circ: &Circuit, seen: &mut HashSet<HashBytes>, f: &mut F)
+    fn recurse<F>(circ: CircuitRc, seen: &mut HashSet<HashBytes>, f: &mut F)
     where
-        F: FnMut(&Circuit),
+        F: FnMut(CircuitRc),
     {
         if !seen.contains(&circ.info().hash) {
             seen.insert(circ.info().hash);
             for child in circ.children() {
-                recurse(&child, seen, f)
+                recurse(child, seen, f)
             }
             f(circ);
         }
@@ -1192,52 +1260,52 @@ where
     recurse(circuit, &mut seen, &mut f);
 }
 
-pub fn deep_map_op<F>(circuit: &Circuit, f: F) -> Option<CircuitRc>
+pub fn deep_map_op<F>(circuit: CircuitRc, f: F) -> Option<CircuitRc>
 where
-    F: Fn(&Circuit) -> Option<CircuitRc>,
+    F: Fn(CircuitRc) -> Option<CircuitRc>,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> Option<CircuitRc> {
-        let inner_mapped = circ.map_children_op(&mut recurse);
+    fn recurse(circ: CircuitRc) -> Option<CircuitRc> {
+        let inner_mapped = circ.map_children_op(&mut recurse).map(|z| z.rc());
         inner_mapped
-            .map(|x| x.and_then_or_clone(&f))
+            .map(|x| f(x.clone()).unwrap_or(x))
             .or_else(|| f(circ))
     }
     recurse(circuit)
 }
 
-pub fn deep_map_pre_new_children<F>(circuit: &Circuit, f: F) -> CircuitRc
+pub fn deep_map_pre_new_children<F>(circuit: CircuitRc, f: F) -> CircuitRc
 where
-    F: Fn(&Circuit, &Vec<CircuitRc>) -> CircuitRc,
+    F: Fn(CircuitRc, &Vec<CircuitRc>) -> CircuitRc,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> CircuitRc {
+    fn recurse(circ: CircuitRc) -> CircuitRc {
         let old_children: Vec<CircuitRc> = circ.children().collect();
-        let new_children = old_children.iter().map(|x| &***x).map(recurse).collect();
+        let new_children = old_children.into_iter().map(recurse).collect();
         f(circ, &new_children)
     }
     recurse(circuit)
 }
 
-pub fn deep_map_op_pre_new_children<F>(circuit: &Circuit, f: F) -> Option<CircuitRc>
+pub fn deep_map_op_pre_new_children<F>(circuit: CircuitRc, f: F) -> Option<CircuitRc>
 where
-    F: Fn(&Circuit, &Vec<CircuitRc>) -> Option<CircuitRc>,
+    F: Fn(CircuitRc, &Vec<CircuitRc>) -> Option<CircuitRc>,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> Option<CircuitRc> {
+    fn recurse(circ: CircuitRc) -> Option<CircuitRc> {
         let old_children: Vec<CircuitRc> = circ.children().collect();
         let new_children: Vec<Option<CircuitRc>> =
-            old_children.iter().map(|x| &***x).map(recurse).collect();
+            old_children.iter().cloned().map(recurse).collect();
         if new_children.iter().all(|x| x.is_none()) {
             f(circ, &old_children)
         } else {
             let new_real_children = zip(old_children, new_children)
                 .map(|(old, new)| new.unwrap_or(old))
                 .collect();
-            Some(f(circ, &new_real_children).unwrap_or_else(|| {
+            Some(f(circ.clone(), &new_real_children).unwrap_or_else(|| {
                 circ.map_children_unwrap_idxs(|i| new_real_children[i].clone())
                     .rc()
             }))
@@ -1247,18 +1315,18 @@ where
 }
 
 pub fn deep_map_fallible_pre_new_children<F>(
-    circuit: &Circuit,
+    circuit: CircuitRc,
     f: F,
 ) -> Result<CircuitRc, CircuitConstructionError>
 where
-    F: Fn(&Circuit, &Vec<CircuitRc>) -> Result<CircuitRc, CircuitConstructionError>,
+    F: Fn(CircuitRc, &Vec<CircuitRc>) -> Result<CircuitRc, CircuitConstructionError>,
 {
     #[apply(cached_lambda)]
     #[key(circ.info().hash, HashBytes)]
-    fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
+    fn recurse(circ: CircuitRc) -> Result<CircuitRc, CircuitConstructionError> {
         let old_children: Vec<CircuitRc> = circ.children().collect(); // need to define this for borrow reasons
         let new_children: Result<Vec<CircuitRc>, CircuitConstructionError> =
-            old_children.iter().map(|x| &***x).map(recurse).collect();
+            old_children.into_iter().map(recurse).collect();
         new_children.and_then(|a| f(circ, &a))
     }
     recurse(circuit)
@@ -1283,39 +1351,39 @@ where
 }
 
 pub fn deep_map_op_context<F, C>(
-    circuit: &Circuit,
+    circuit: CircuitRc,
     f: &F,
     context: &mut C,
     self_cache: &mut HashMap<HashBytes, Option<CircuitRc>>,
 ) -> Option<CircuitRc>
 where
-    F: Fn(&Circuit, &mut C) -> Option<CircuitRc>,
+    F: Fn(CircuitRc, &mut C) -> Option<CircuitRc>,
 {
     if let Some(z) = self_cache.get(&circuit.info().hash) {
         return z.clone();
     }
     let inner_mapped = circuit.map_children_op(|x| deep_map_op_context(x, f, context, self_cache));
     let result = match inner_mapped {
-        Some(z) => f(&z, context).or(Some(z.rc())),
-        None => f(circuit, context),
+        Some(z) => f(z.clone().rc(), context).or(Some(z.rc())),
+        None => f(circuit.clone(), context),
     };
     self_cache.insert(circuit.info().hash, result.clone());
     result
 }
 
 pub fn deep_map_op_context_preorder_stoppable<F, C>(
-    circuit: &Circuit,
+    circuit: CircuitRc,
     f: &F,
     context: &mut C,
     self_cache: &mut HashMap<HashBytes, Option<CircuitRc>>,
 ) -> Option<CircuitRc>
 where
-    F: Fn(&Circuit, &mut C) -> (Option<CircuitRc>, bool),
+    F: Fn(CircuitRc, &mut C) -> (Option<CircuitRc>, bool),
 {
     if let Some(z) = self_cache.get(&circuit.info().hash) {
         return z.clone();
     }
-    let (circuit_applied, stop) = f(circuit, context);
+    let (circuit_applied, stop) = f(circuit.clone(), context);
     if stop {
         return circuit_applied;
     }
@@ -1338,12 +1406,12 @@ where
 }
 
 pub fn evaluate_fn_uncached(
-    circ: &Circuit,
+    circ: CircuitRc,
     device_dtype: &TorchDeviceDtype,
 ) -> Result<Tensor, TensorEvalError> {
     let child_tensors: Result<Vec<Tensor>, TensorEvalError> = circ
         .children()
-        .map(|x| evaluate_fn_uncached(&x, device_dtype))
+        .map(|x| evaluate_fn_uncached(x, device_dtype))
         .collect();
     let child_tensors = child_tensors?;
 
@@ -1551,7 +1619,7 @@ impl PyCircuitBase {
 
     #[args(device_dtype = "Default::default()")]
     fn evaluate(&self, device_dtype: TorchDeviceDtypeOp) -> Result<Tensor, TensorEvalError> {
-        evaluate_fn_dtype_device(self, device_dtype)
+        evaluate_fn_dtype_device((***self).clone().rc(), device_dtype)
     }
 }
 
