@@ -28,10 +28,10 @@ use uuid::{uuid, Uuid};
 #[pyfunction]
 #[pyo3(name = "expand_node")]
 pub fn expand_node_py(circuit: CircuitRc, inputs: Vec<CircuitRc>) -> CircResult {
-    expand_node(&circuit, &inputs)
+    expand_node(circuit, &inputs)
 }
 
-pub fn expand_node(circuit: &Circuit, inputs: &Vec<CircuitRc>) -> CircResult {
+pub fn expand_node(circuit: CircuitRc, inputs: &Vec<CircuitRc>) -> CircResult {
     if inputs.len() != circuit.children().count() {
         return Err(CircuitConstructionError::ModuleNodeWrongNumberChildren {
             expected: circuit.children().count(),
@@ -50,8 +50,8 @@ pub fn expand_node(circuit: &Circuit, inputs: &Vec<CircuitRc>) -> CircResult {
     let batch_shapes: Vec<&[usize]> = zip(&batch_ranks, inputs)
         .map(|(br, new)| &new.info().shape[0..*br])
         .collect();
-    match circuit {
-        Circuit::Symbol(_) | Circuit::ScalarConstant(_) => Ok(circuit.clone().rc()),
+    match &**circuit {
+        Circuit::Symbol(_) | Circuit::ScalarConstant(_) => Ok(circuit.clone()),
         Circuit::Rearrange(node) => {
             let input_shape_non_batch = inputs[0].info().shape[batch_ranks[0]..]
                 .iter()
@@ -131,7 +131,7 @@ pub fn expand_node(circuit: &Circuit, inputs: &Vec<CircuitRc>) -> CircResult {
                         .iter()
                         .map(|x| x.iter().cloned().collect())
                         .collect(),
-                    circuit: circuit.clone().rc(),
+                    circuit: circuit.clone(),
                 });
             }
             if !zip(&node.nodes, zip(inputs, &batch_ranks)).all(|(old, (new, br))| {
@@ -166,7 +166,7 @@ pub fn expand_node(circuit: &Circuit, inputs: &Vec<CircuitRc>) -> CircResult {
                                     .iter()
                                     .map(|x| x.iter().cloned().collect())
                                     .collect(),
-                                circuit: circuit.clone().rc(),
+                                circuit: circuit.clone(),
                             });
                         }
                     } else {
@@ -206,7 +206,7 @@ pub fn expand_node(circuit: &Circuit, inputs: &Vec<CircuitRc>) -> CircResult {
         }
         _ => {
             if inputs[..] == circuit.children().collect::<Vec<_>>()[..] {
-                Ok(circuit.clone().rc())
+                Ok(circuit.clone())
             } else {
                 Err(CircuitConstructionError::ExpandNodeUnknownVariant {
                     variant: circuit.variant_string(),
@@ -222,13 +222,13 @@ pub fn replace_expand_bottom_up_dict_py(
     circuit: CircuitRc,
     dict: HashMap<CircuitRc, CircuitRc>,
 ) -> CircResult {
-    replace_expand_bottom_up(circuit, |x| dict.get(&x.clone().rc()).cloned())
+    replace_expand_bottom_up(circuit, |x| dict.get(&x).cloned())
 }
 
 #[pyfunction]
 #[pyo3(name = "replace_expand_bottom_up")]
 pub fn replace_expand_bottom_up_py(circuit: CircuitRc, f: PyObject) -> CircResult {
-    replace_expand_bottom_up(circuit, |x| pycall!(f, (x.clone().rc(),)))
+    replace_expand_bottom_up(circuit, |x| pycall!(f, (x.clone(),)))
 }
 
 pub fn replace_expand_bottom_up<F>(
@@ -236,17 +236,17 @@ pub fn replace_expand_bottom_up<F>(
     replacer: F,
 ) -> Result<CircuitRc, CircuitConstructionError>
 where
-    F: Fn(&Circuit) -> Option<CircuitRc>,
+    F: Fn(CircuitRc) -> Option<CircuitRc>,
 {
-    let recursor = |circuit: &Circuit,
+    let recursor = |circuit: CircuitRc,
                     new_children: &Vec<CircuitRc>|
      -> Result<CircuitRc, CircuitConstructionError> {
-        if let Some(replaced) = replacer(circuit) {
+        if let Some(replaced) = replacer(circuit.clone()) {
             return Ok(replaced);
         }
         expand_node(circuit, new_children)
     };
-    deep_map_fallible_pre_new_children(&circuit, recursor)
+    deep_map_fallible_pre_new_children(circuit, recursor)
 }
 
 #[pyclass(unsendable)]
@@ -320,7 +320,7 @@ impl ModuleNodeSpec {
         replace_expand_bottom_up(self.spec_circuit.clone(), |c| {
             self.input_specs
                 .iter()
-                .position(|x| &x.symbol.clone().c() == c)
+                .position(|x| x.symbol.info().hash == c.info().hash)
                 .map(|i| nodes[i].clone())
         })
     }
@@ -417,8 +417,8 @@ impl ModuleNodeSpec {
     ) -> Result<Self, CircuitConstructionError> {
         let mut input_specs_dict: HashMap<usize, ModuleNodeArgSpec> = HashMap::new();
         let circuit = deep_map_op_context(
-            &circuit,
-            &|sub: &Circuit, input_specs_dict: &mut HashMap<usize, ModuleNodeArgSpec>| {
+            circuit.clone(),
+            &|sub: CircuitRc, input_specs_dict: &mut HashMap<usize, ModuleNodeArgSpec>| {
                 if let Some(sym) = sub.as_symbol() {
                     static RE_NUM_BEGIN: Lazy<Regex> =
                         Lazy::new(|| Regex::new(r"#(\d+) (.*)").unwrap());
@@ -464,7 +464,7 @@ impl ModuleNodeSpec {
     ) -> Option<Self> {
         let mut new_input_specs: Vec<Option<ModuleNodeArgSpec>> = vec![None; input_specs.len()];
         let spec_circuit = deep_map_op_context_preorder_stoppable(
-            &circuit,
+            circuit,
             &|circuit,
               c: &mut (
                 &mut Vec<Option<ModuleNodeArgSpec>>,
@@ -644,13 +644,13 @@ impl CircuitNode for ModuleNode {
     ) -> Result<Self, CircuitConstructionError>
     where
         CircuitConstructionError: From<E>,
-        F: FnMut(usize, &'a Circuit) -> Result<CircuitRc, E>,
+        F: FnMut(usize, CircuitRc) -> Result<CircuitRc, E>,
     {
         Self::try_new(
             self.nodes
                 .iter()
                 .enumerate()
-                .map(move |(i, circ)| f(i, circ))
+                .map(move |(i, circ)| f(i, circ.clone()))
                 .collect::<Result<Vec<_>, _>>()?,
             self.spec.clone(),
             self.name.clone(),
@@ -731,9 +731,9 @@ impl ModuleNode {
 
 #[pyfunction]
 pub fn inline_all_modules(circuit: CircuitRc) -> CircuitRc {
-    deep_map_unwrap(&circuit, |c| match c {
+    deep_map_unwrap(circuit, |c| match &**c {
         Circuit::ModuleNode(mn) => inline_all_modules(mn.expand()),
-        _ => c.clone().rc(),
+        _ => c.clone(),
     })
 }
 use std::cell::RefCell;
