@@ -4,12 +4,16 @@ use super::{
 };
 use super::{CircuitConstructionError, CircuitRc};
 use crate::all_imports::{Shape, Sv, TensorIndex};
-use crate::circuit::{CircResult, CircuitNode, ModuleNode, ModuleNodeSpec};
+use crate::circuit::{
+    AutoTag, CircResult, CircuitNode, Cumulant, DiscreteVar, ModuleNode, ModuleNodeSpec,
+    StoredCumulantVar,
+};
 use crate::hashmaps::FxHashMap as HashMap;
 use crate::pyo3_prelude::*;
 use crate::tensor_util::TorchDeviceDtypeOp;
 use once_cell::sync::Lazy;
 use regex::Regex;
+use std::iter::zip;
 use std::rc::Rc;
 use std::str::FromStr;
 use uuid::Uuid;
@@ -52,11 +56,25 @@ pub fn parse_compiler_repr_bijection(
         let re_captures = RE.captures(line).ok_or_else(rf)?;
         let num_spaces = re_captures.get(1).ok_or_else(rf)?.as_str().len();
         if num_spaces % tab_width != 0 {
-            return Err(CircuitConstructionError::ParsingInvalidIndentation { spaces: num_spaces });
+            return Err(CircuitConstructionError::ParsingInvalidIndentation {
+                tab_width: tab_width,
+                spaces: num_spaces,
+                stack_indentation: stack.len(),
+                stack_top: stack
+                    .last()
+                    .map(|z| partial_circuits[*z].clone().unwrap().variant.clone()),
+            });
         }
         let indentation_level = num_spaces / tab_width;
         if indentation_level > stack.len() {
-            return Err(CircuitConstructionError::ParsingInvalidIndentation { spaces: num_spaces });
+            return Err(CircuitConstructionError::ParsingInvalidIndentation {
+                tab_width: tab_width,
+                spaces: num_spaces,
+                stack_indentation: stack.len(),
+                stack_top: stack
+                    .last()
+                    .map(|z| partial_circuits[*z].clone().unwrap().variant.clone()),
+            });
         }
         stack.truncate(indentation_level);
         let serial_number = re_captures
@@ -159,7 +177,7 @@ pub fn parse_compiler_repr_bijection(
                                 variant: variant.to_owned(),
                             })?,
                         ps.name.clone(),
-                        tensors_as_random_device_dtype.clone(),
+                        tensors_as_random_device_dtype,
                     )
                     .rc())
                 } else {
@@ -269,6 +287,53 @@ pub fn parse_compiler_repr_bijection(
                 .ok_or_else(|| CircuitConstructionError::CircuitRefNotFound {
                     name: ps.extra.clone(),
                 }),
+            "AutoTag" => Uuid::from_str(&ps.extra)
+                .map_err(|_| CircuitConstructionError::InvalidUuid {
+                    string: ps.extra.clone(),
+                })
+                .map(|uuid| AutoTag::nrc(children[0].clone(), uuid, ps.name.clone())),
+            "Cumulant" => Ok(Cumulant::nrc(children, ps.name.clone())),
+            "DiscreteVar" => {
+                DiscreteVar::try_new(children[0].clone(), children[1].clone(), ps.name.clone())
+                    .map(|z| z.rc())
+            }
+            "StoredCumulantVar" => {
+                if let [cum_nums, uuid] = ps.extra.split("|").collect::<Vec<_>>()[..] {
+                    cum_nums
+                        .split(",")
+                        .map(|s| s.trim().parse::<usize>())
+                        .collect::<Result<Vec<_>, _>>()
+                        .map_err(|e| CircuitConstructionError::ParsingNumberFail {
+                            string: format!("{}", e),
+                        })
+                        .and_then(|keys| {
+                            Uuid::from_str(uuid)
+                                .map_err(|_| CircuitConstructionError::InvalidUuid {
+                                    string: ps.extra.clone(),
+                                })
+                                .map(|uuid| (keys, uuid))
+                        })
+                        .and_then(|(keys, uuid)| {
+                            if keys.len() != children.len() {
+                                Err(CircuitConstructionError::ParsingWrongNumberChildren {
+                                    expected: keys.len(),
+                                    found: children.len(),
+                                })
+                            } else {
+                                StoredCumulantVar::try_new(
+                                    zip(keys, children).collect(),
+                                    uuid,
+                                    ps.name.clone(),
+                                )
+                                .map(|z| z.rc())
+                            }
+                        })
+                } else {
+                    Err(CircuitConstructionError::ParsingFail {
+                        string: ps.extra.clone(),
+                    })
+                }
+            }
             _ => Err(CircuitConstructionError::ParsingInvalidVariant {
                 v: variant.to_owned(),
             }),
