@@ -671,10 +671,10 @@ impl Dag {
     }
 }
 
-pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
+pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
     let mut result: Dag = Default::default();
     // append the circuit to the dag if it's not already there, and return its index
-    let number_node = |c: &Circuit, result: &mut Dag| -> usize {
+    let number_node = |c: CircuitRc, result: &mut Dag| -> usize {
         if let Some(idx) = result.hash_to_node.get(&c.info().hash) {
             return *idx;
         }
@@ -688,18 +688,18 @@ pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
         result.node_hashes.len() - 1
     };
     for circuit in circuits {
-        visit_circuit(circuit, &mut |c: &Circuit| {
+        visit_circuit(circuit.clone(), &mut |c: CircuitRc| {
             // ArrayConstants are never added to the dag to begin with because they're always 0 cost
-            if !Leaf::matches(c) {
-                let my_number = number_node(c, &mut result);
+            if !Leaf::matches(&c) {
+                let my_number = number_node(c.clone(), &mut result);
                 let children_to_consider: Vec<CircuitRc> =
                     c.children().filter(|child| !Leaf::matches(child)).collect();
                 result.children[my_number] = children_to_consider
                     .iter()
-                    .map(|child| number_node(child, &mut result))
+                    .map(|child| number_node(child.clone(), &mut result))
                     .collect();
                 for child in children_to_consider {
-                    let new_num = number_node(&child, &mut result);
+                    let new_num = number_node(child.clone(), &mut result);
                     result.parents[new_num].push(my_number);
                 }
             }
@@ -806,7 +806,7 @@ pub fn circuit_to_schedule(
     circuit: CircuitRc,
     context: &mut OptimizationContext,
 ) -> Result<Schedule, SchedulingError> {
-    let dag = circuits_to_dag(&[&**circuit]);
+    let dag = circuits_to_dag(&[circuit.clone()]);
     if circuit.info().max_non_input_size > BigUint::from(context.cache.fallback_total_numel) {
         return Err(SchedulingError::SingleElementTooBig {
             numel: circuit.info().max_non_input_size.clone(),
@@ -863,7 +863,7 @@ pub fn circuit_to_schedule(
 
     let order = order_result?;
 
-    let to_node = hash_to_node(&circuit);
+    let to_node = hash_to_node(circuit.clone());
     let circuit_order: Vec<CircuitRc> = order
         .iter()
         .map(|x| to_node[&dag.node_hashes[*x]].clone())
@@ -1004,7 +1004,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                                 Instruction::Compute(inner_k, c) => {
                                     let new_c = c
                                         .map_children_unwrap(|child| {
-                                            let child_inner_k = get_child_key(child.clone().rc());
+                                            let child_inner_k = get_child_key(child.clone());
                                             let child_outer_k =
                                                 inner_to_outer_key.get(&child_inner_k).unwrap();
                                             child_from_key(
