@@ -48,6 +48,7 @@ pub mod opt_einsum;
 pub mod py_types;
 pub mod rearrange_spec;
 pub mod rrfs;
+pub mod set_cover;
 pub mod smallvec;
 pub mod tensor_util;
 mod union_find;
@@ -135,6 +136,17 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit::module_nodes::ModuleNode>()?;
     m.add_class::<circuit::module_nodes::ModuleNodeSpec>()?;
     m.add_class::<circuit::module_nodes::ModuleNodeArgSpec>()?;
+    m.add_class::<circuit::module_nodes::ModuleNodeArgSpec>()?;
+    m.add_class::<circuit::AutoTag>()?;
+    m.add_class::<circuit::DiscreteVar>()?;
+    m.add_class::<circuit::StoredCumulantVar>()?;
+    m.add_class::<circuit::Cumulant>()?;
+    m.add_class::<circuit::circuit_matchers::Matcher>()?;
+
+    m.add_function(wrap_pyfunction!(circuit::circuit_is_leaf, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::circuit_is_irreducible_node, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::circuit_is_constant, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::circuit_is_var, m)?)?;
 
     m.add_function(wrap_pyfunction!(add_collapse_scalar_inputs, m)?)?;
     m.add_function(wrap_pyfunction!(add_deduplicate, m)?)?;
@@ -185,6 +197,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(canonicalize_node_py, m)?)?;
 
     m.add_function(wrap_pyfunction!(circuit::batching::batch_to_concat, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::batching::batch_einsum_py, m)?)?;
     m.add_function(wrap_pyfunction!(
         circuit::compiler_strip::strip_names_py,
         m
@@ -233,6 +246,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_function(wrap_pyfunction!(flat_concat, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit::flat_concat_back, m)?)?;
     m.add_function(wrap_pyfunction!(deep_heuristic_nest_adds, m)?)?;
     m.add_function(wrap_pyfunction!(concat_fuse, m)?)?;
     m.add_function(wrap_pyfunction!(generalfunction_pull_concat, m)?)?;
